==> accessible-pipeline-0.5.0.tgz: package/src/run.ts <==
import {AxePuppeteer} from 'axe-puppeteer';
import puppeteer, {Browser, Page} from 'puppeteer';
import {AxeResults} from 'axe-core';
import cheerio from 'cheerio';
import fs from 'fs';
import path from 'path';
import {promisify} from 'util';
import {logger} from './logger';
import matchit from 'matchit';

// Create a child logger scoped to the module
const log = logger.child({module: 'root'});

// Create a logger for results. This is the one that gets consumed by streaming reporters.
const resultLog = logger.child({module: 'results'});

// Promise-friendly core fns
const writeFile = promisify(fs.writeFile);
const readFile = promisify(fs.readFile);

// TODO: Source-map-support
// TODO: Add non-crawl option
// TODO: Add multiple roots
// TODO: For multiple roots and crawling, what do we do?

// TODO: more granular error boundaries, and error reporting
// for example {..., fetcherErrors: [{url: ..., reason: ...}]}
// - "Friends don't let friends not handle errors"

// TODO: Consider AxE reporting verbosity toggle
// TODO: Ability to set a UA to hide from analytics

// Common options for excluding links to visit
// For example *.pdf and #heading-link came up often in early iterations
type Options = {
  /* The maximum number of pages to visit */
  pageLimit: number;
  /* The maximum number of retries for a failing page */
  maxRetries?: number;
  /* Whether to ignore links of the shape https://example.com#my-id */
  ignoreFragmentLinks?: boolean;
  /* A list of extensions to ignore, skipping pages */
  ignoreExtensions?: Array<string>;
  /* A path to a route manifest file, used to de-duplicate visited pages and routes */
  routeManifestPath?: string;
  /* Whether to expose the streaming logging API, used for advanced, "live" reporters */
  streaming?: boolean;
};

/**
 * The core function that runs assertions on the provided URLs, but does not touch the filesystem.
 * Used as the main export of the module, for programatic use.
 */
export async function runCore(rootURL: URL, opts: Options) {
  // Load RouteManifest, if specified
  // TODO: Consider a union for this
  let ROUTE_MANIFEST;
  if (opts.routeManifestPath) {
    // Read the manifest specified
    try {
      ROUTE_MANIFEST = await readFile(opts.routeManifestPath, 'utf8').then(
        data => JSON.parse(data)
      );
    } catch (err) {
      log.error('There was an error when trying to read the route manifest');
      throw err;
    }
    log.info('Read manifest!');
    log.debug(JSON.stringify(ROUTE_MANIFEST, null, 2));
    // TODO: Validate the format
  }

  // TODO: Make a "pages considered" list as well
  const PAGES_VISITED = new Set<string>();
  const ROUTES_VISITED = new Set<string>();

  // It is preferable to store a string in the set, otherwise identities get jumbled with objects...
  // Wish there was a good hashed set implementation
  // To compensate, we store the string, and transform to/from URL href at the edges
  let PAGES_TO_VISIT = new Set([rootURL.href]);
  let RESULTS: Array<AxeResults> = [];

  log.info('Will run with:', {...opts});

  const browser = await puppeteer.launch();
  let run = 0;

  // TODO: Consider Depth-First Search vs. Breadth-First Search
  for (const pageHref of PAGES_TO_VISIT) {
    // First, convert to a URL (if we have gotten this from the scripts below, this is OK)
    const pageUrl = new URL(pageHref);

    // Double check that the page has not been visited (might come into play for concurrency)
    const {shouldProcess: shouldRun, reason} = shouldProcess(
      ROUTE_MANIFEST,
      ROUTES_VISITED,
      PAGES_VISITED,
      opts.pageLimit,
      run,
      pageUrl
    );

    log.info(
      `Page: ${pageUrl.href}. Should process: ${shouldRun}, reason: ${
        reason.type
      }`
    );

    // Remove from queue if it will not run
    if (!shouldRun) {
      PAGES_TO_VISIT.delete(pageUrl.href);
    }

    if (shouldRun) {
      run++;
      log.trace('Run', run);

      log.info('Will process', pageUrl.href);

      // Also add "in progress" to the result log
      streamingSendInProgress(pageUrl.href, opts.streaming);

      // TODO: streamingSendReportError(streaming, {href: pageUrl.href, error: err});

      // Add to pages visited, remove from queue
      PAGES_VISITED.add(pageUrl.href);
      PAGES_TO_VISIT.delete(pageUrl.href);

      // If the reason we selected a page was because of a route, then add that route to the visited set
      if (reason.type === 'Route') {
        ROUTES_VISITED.add(reason.route);
      }

      // Process the page, get results
      const attemptRun = async () => {
        let succeeded = false;
        for (let tryCount = 1; tryCount <= opts.maxRetries!; tryCount++) {
          if (!succeeded) {
            log.trace(`Attempt ${tryCount} of ${opts.maxRetries}`);
            try {
              const processResults = await processPage(browser, pageUrl);
              succeeded = true;
              return processResults;
            } catch (err) {
              log.error('There was an error processing the page', err);
            }
          }
        }

        if (!succeeded) {
          // TODO: streamingSendErrorProcessing(pageUrl.href, error)
          return Promise.resolve({results: [], nextPages: []});
        }
      };

      const runAttempt = await attemptRun();
      let results = runAttempt!.results as AxeResults;
      let nextPages = runAttempt!.nextPages;

      // Append the results to the list
      // TODO: Could instead opt to return the results as a stream, and decide where to write in the consumer
      RESULTS = RESULTS.concat(results);

      // Send a 'GotResults' message on the streaming channel
      streamingSendGotResults(results, opts.streaming);

      // Remove any visited pages, and add the rest to the ones to visit
      const pagesToVisit = getPagesToVisit(nextPages, PAGES_VISITED, opts);

      // NOTE: We trasnform from a URL to href for storage, see the reasons in Set above
      pagesToVisit.forEach(page => PAGES_TO_VISIT.add(page.href));
    }
  }

  // Close the browser and return reports and state
  await browser.close();

  // Write serialized state
  const stateObj = {
    entry: rootURL.href,
    options: opts,
    run: run,
    routes: ROUTE_MANIFEST,
    pagesVisited: Array.from(PAGES_VISITED.keys()),
    routesVisited: Array.from(ROUTES_VISITED.keys()),
    toVisit: Array.from(PAGES_TO_VISIT.keys()),
  };

  return {results: RESULTS, state: stateObj};
}

async function main(rootURL: URL, opts: Options) {
  // Run the core functionality
  const {results, state} = await runCore(rootURL, opts);

  // Write reports and state

  // Generate some sort of id for the run
  const runId = Date.now();
  const reportFileName = `report-${runId}.json`;
  const stateFileName = `state-${runId}.json`;

  // Write report
  await writeFile(reportFileName, JSON.stringify(results, null, 2), 'utf8');
  log.info(`Wrote ${reportFileName}`);
  streamingSendWroteReportFile(reportFileName, opts.streaming);

  await writeFile(stateFileName, JSON.stringify(state, null, 2), 'utf8');
  log.info(`Wrote ${stateFileName}`);
  streamingSendWroteStateFile(stateFileName, opts.streaming);
}

async function processPage(browser: Browser, pageUrl: URL) {
  const page = await browser.newPage();
  await page.setBypassCSP(true);

  // Use 'networkidle2' to allow time for http-equiv redirects etc.
  const response = await page.goto(pageUrl.href, {waitUntil: 'networkidle2'});
  log.info(pageUrl.href);

  // Analyse page, get results
  const results = await analysePage(page);

  // Gather next links
  const nextPages = await gatherNextPages(pageUrl, page);

  await page.close();
  return {results, nextPages};
}

async function analysePage(pageObj: Page) {
  return await new AxePuppeteer(pageObj).analyze();
}

async function gatherNextPages(currentPageUrl: URL, pageObj: Page) {
  // Gather all links on a page
  const htmlContent = await pageObj.content();
  const $ = cheerio.load(htmlContent);
  const links = $(`a`);
  const linkUrls = links.map((i, link) => link.attribs['href']).get();

  // Convert a URL to an absolute url, if it is relative, then filter by origin
  /*
   * @example new URL('/en/page', 'https://example.com').href === 'https://example.com/en/page
   * @example new URL('/en/page', 'https://example.com/otherpage').href === 'https://example.com/en/page
   * @example new URL('https://otherexample.com/en/page', 'https://example.com').href !== 'https://example.com/en/page
   */
  const relevantUrls = linkUrls
    .map(url => new URL(url, currentPageUrl))
    .filter(url => url.origin === currentPageUrl.origin);

  return relevantUrls;
}

function getPagesToVisit(
  urls: Array<URL>,
  visited: Set<string>,
  options: Options
) {
  // TODO: We could some fancy things here, such as:
  //  - having a maximumDepth counter
  //  - some other metric / heuristic?

  return urls
    .filter(url => !visited.has(url.href))
    .filter(url => {
      if (options.ignoreExtensions) {
        const ext = path.extname(url.pathname);
        // Reject if extension is in ignore list
        return !options.ignoreExtensions.includes(ext);
      }
      // Pass through otherwise
      return true;
    })
    .filter(url => {
      if (options.ignoreFragmentLinks) {
        // Reject if there is a fragment
        return !url.hash;
      }
      // Pass through otherwise
      return true;
    });
}

//
// PROCESSING

type ProcessDecision = {
  shouldProcess: boolean;
  reason: ProcessReason;
};

type ProcessReason =
  | {type: 'PageLimit'}
  | {type: 'Verbatim'; href: string}
  | {type: 'Route'; route: string};

const ReasonPageLimit = (): ProcessReason => ({type: 'PageLimit'});

const ReasonVerbatim = (href: string): ProcessReason => ({
  type: 'Verbatim',
  href,
});

const ReasonRoute = (route: string): ProcessReason => ({type: 'Route', route});

/** Decide whether a pageHref should be visited */
function shouldProcess(
  ROUTE_MANIFEST: Array<string>,
  ROUTES_VISITED: Set<string>,
  PAGES_VISITED: Set<string>,
  PAGE_LIMIT: number,
  run: number,
  pageUrl: URL
): ProcessDecision {
  // TODO: Also add things to the relevant _VISITED page!
  if (run >= PAGE_LIMIT) {
    log.debug('Page limit reached');
    return {shouldProcess: false, reason: ReasonPageLimit()};
  } else {
    // If the route manifest is specified, consider the route
    if (ROUTE_MANIFEST) {
      const routes = ROUTE_MANIFEST.map(matchit.parse);
      const matchesRoute = (str: string) => matchit.match(str, routes);
      const routeMatch = matchesRoute(pageUrl.pathname);

      if (routeMatch.length) {
        // If the page matches one of the routes specified, then check if visited
        const {old: matchedRoute} = routeMatch[0];
        log.debug('Route matches, will check if visited');
        return {
          shouldProcess: !ROUTES_VISITED.has(matchedRoute),
          reason: ReasonRoute(matchedRoute),
        };
      } else {
        // If no match, then consider the verbatim version
        log.debug('No route matches, will check verbatim');
        return {
          shouldProcess: !PAGES_VISITED.has(pageUrl.href),
          reason: ReasonVerbatim(pageUrl.href),
        };
      }
    } else {
      // If no manifest, then consider the verbatim version
      log.debug('No manifest, will check verbatim');
      return {
        shouldProcess: !PAGES_VISITED.has(pageUrl.href),
        reason: ReasonVerbatim(pageUrl.href),
      };
    }
  }
}

//
// STREAMING

function streamingSendInProgress(href: string, streaming?: boolean) {
  if (streaming) {
    resultLog.info({msg: 'InProgress', href});
  }
}

function streamingSendWroteReportFile(filename: string, streaming?: boolean) {
  if (streaming) {
    resultLog.info({
      msg: 'DisplayInfo',
      description: `Wrote the report to ${filename}. You can view it again with\n\`$ accessible-pipeline view --file ${filename}\`.`,
    });
  }
}

function streamingSendWroteStateFile(filename: string, streaming?: boolean) {
  if (streaming) {
    resultLog.info({
      msg: 'DisplayInfo',
      description: `Wrote the state to ${filename}. It might me useful if you want to debug which pages were visited.`,
    });
  }
}

function streamingSendGotResults(results: AxeResults, streaming?: boolean) {
  // Write to the results stream, after rewriting results to only keep the node targets
  // Helps avoid some odd edge cases where the JSON breaks parsing when piping, because of the HTML content
  // TODO: Declare StreamingAxeResults type, which matches the shape we actually send.
  if (streaming) {
    resultLog.info({
      msg: 'GotResults',
      results: {
        ...results,
        violations: results.violations.map(violation => ({
          ...violation,
          nodes: violation.nodes.map(node => ({
            // Only keep the target, since that's what we report on
            target: node.target,
          })),
        })),
        passes: results.passes.map(pass => ({
          ...pass,
          nodes: pass.nodes.map(node => ({
            // Only keep the target, since that's what we report on
            target: node.target,
          })),
        })),
        incomplete: results.incomplete.map(violation => ({
          ...violation,
          nodes: violation.nodes.map(node => ({
            // Only keep the target, since that's what we report on
            target: node.target,
          })),
        })),
        inapplicable: results.inapplicable.map(violation => ({
          ...violation,
          nodes: violation.nodes.map(node => ({
            // Only keep the target, since that's what we report on
            target: node.target,
          })),
        })),
      },
    });
  }
}

//
// CLI

export type CLIOptions = Options & {
  streaming: boolean;
  ci: boolean;
};

// TODO: Accept multiple roots in the future
export function cliEntry(rootHref: string, opts: CLIOptions) {
  // Whether to output a results stream under resultsLog
  const {ci, ...options} = opts;

  // Read the url as the first CLI argument
  let rootURL: URL;

  try {
    rootURL = new URL(rootHref);
  } catch (err) {
    log.fatal(
      'The URL you provided was in an unexpected format: ',
      err.toString()
    );
    process.exit(1);
  }

  main(rootURL!, options).catch(err => {
    // TODO: Consider other ways of reporting errors here
    log.fatal('The process encountered an unrecoverable error: ', err);
    process.exit(1);
  });
}

==> accessible-pipeline-0.6.0.tgz: package/src/run.ts <==
import fs from 'fs';
import {promisify} from 'util';
import * as core from './core';
import * as streaming from './streaming';
import {logger} from './logger';

// Promise-friendly core fns
const writeFile = promisify(fs.writeFile);

// Create a child logger scoped to the module
const log = logger.child({module: 'run'});

/** Code backing the "run" CLI command */
async function main(rootURL: URL, opts: core.Options) {
  // Run the core functionality
  const {results, state} = await core.runCore(rootURL, opts);

  // Write reports and state

  // Generate some sort of id for the run
  const runId = Date.now();
  const reportFileName = `report-${runId}.json`;
  const stateFileName = `state-${runId}.json`;

  // Write report
  await writeFile(reportFileName, JSON.stringify(results, null, 2), 'utf8');
  log.info(`Wrote ${reportFileName}`);
  streaming.sendWroteReportFile(reportFileName, opts.streaming);

  await writeFile(stateFileName, JSON.stringify(state, null, 2), 'utf8');
  log.info(`Wrote ${stateFileName}`);
  streaming.sendWroteStateFile(stateFileName, opts.streaming);
}

//
// CLI

export type CLIOptions = core.Options & {
  streaming: boolean;
  ci: boolean;
};

// TODO: Accept multiple roots in the future
export function cliEntry(rootHref: string, opts: CLIOptions) {
  // Whether to output a results stream under resultsLog
  const {ci, ...options} = opts;

  // Read the url as the first CLI argument
  let rootURL: URL;

  try {
    rootURL = new URL(rootHref);
  } catch (err) {
    log.fatal(
      'The URL you provided was in an unexpected format: ',
      err.toString()
    );
    process.exit(1);
  }

  main(rootURL!, options).catch(err => {
    // TODO: Consider other ways of reporting errors here
    log.fatal('The process encountered an unrecoverable error: ', err);
    process.exit(1);
  });
}
